from dataclasses import dataclass, field
from typing import (
    Any,
    Awaitable,
    Callable,
    List,
    MutableMapping,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)


@dataclass
class RouteMapping:
    route: str = ""
    func: Optional[Callable] = None
    module_name: Optional[str] = None
    qual_name: Optional[str] = None
    http_method: Optional[str] = None

    def __bool__(self) -> bool:
        return (
            True
            if self.route
            and self.func
            and self.module_name
            and self.qual_name
            and self.http_method
            else False
        )

    def __hash__(self) -> int:
        return (
            hash(self.route)
            ^ hash(self.http_method)
            ^ hash(self.module_name)
            ^ hash(self.qual_name)
        )


@dataclass
class RouteMappings:
    mappings: List[RouteMapping] = field(default_factory=list)
    http_method: str = "GET"

    def get_map_by_func(self, func_name: str, parent_name: str = None) -> RouteMapping:
        for mapping in self.mappings:
            if mapping.func and mapping.func.__name__ == func_name:
                if parent_name:
                    if mapping.qual_name == parent_name:
                        return mapping
                else:
                    return mapping
        return RouteMapping()

    def get_map_by_route(self, route: str, parent_name: str = None) -> RouteMapping:
        for mapping in self.mappings:
            if mapping.route == route:
                if parent_name:
                    if mapping.qual_name == parent_name:
                        return mapping
                else:
                    return mapping
        return RouteMapping()


@dataclass
class Methods:
    GET: RouteMappings
    POST: RouteMappings
    PUT: RouteMappings
    DELETE: RouteMappings
    PATCH: RouteMappings
    HEAD: RouteMappings
    OPTIONS: RouteMappings
    WEBSOCKET: RouteMappings

    def __init__(self) -> None:
        self.GET = RouteMappings(http_method="GET")
        self.POST = RouteMappings(http_method="POST")
        self.PUT = RouteMappings(http_method="PUT")
        self.DELETE = RouteMappings(http_method="DELETE")
        self.PATCH = RouteMappings(http_method="PATCH")
        self.HEAD = RouteMappings(http_method="HEAD")
        self.OPTIONS = RouteMappings(http_method="OPTIONS")
        self.WEBSOCKET = RouteMappings(http_method="WEBSOCKET")

    def get(self, http_method: str) -> RouteMappings:
        if hasattr(self, http_method):
            return getattr(self, http_method)
        else:
            raise TypeError(f"{http_method} is not a supported HTTP Method")

    def get_all_maps_by_func(
        self, func_name: str, parent_name: str = None
    ) -> Set[RouteMapping]:
        return {
            self.GET.get_map_by_func(func_name, parent_name),
            self.POST.get_map_by_func(func_name, parent_name),
            self.PUT.get_map_by_func(func_name, parent_name),
            self.PATCH.get_map_by_func(func_name, parent_name),
            self.DELETE.get_map_by_func(func_name, parent_name),
            self.HEAD.get_map_by_func(func_name, parent_name),
            self.OPTIONS.get_map_by_func(func_name, parent_name),
            self.WEBSOCKET.get_map_by_func(func_name, parent_name),
        }

    def get_all_maps_by_route(
        self, route: str, parent_name: str = None
    ) -> Set[RouteMapping]:
        return {
            self.GET.get_map_by_route(route, parent_name),
            self.POST.get_map_by_route(route, parent_name),
            self.PUT.get_map_by_route(route, parent_name),
            self.PATCH.get_map_by_route(route, parent_name),
            self.DELETE.get_map_by_route(route, parent_name),
            self.HEAD.get_map_by_route(route, parent_name),
            self.OPTIONS.get_map_by_route(route, parent_name),
            self.WEBSOCKET.get_map_by_route(route, parent_name),
        }


class Node:
    def __init__(self, route: str, class_instance: Type):
        self.route = route
        self.class_instance = class_instance
        self.parent: Optional[Branch] = None


class Branch(Node):
    def __init__(
        self, route: str, class_instance: Type, children: List["TreePart"] = None
    ):
        super().__init__(route, class_instance)
        self.children = children or []


class Leaf(Node):
    def __init__(self, route: str, class_instance: Type, route_mapping: RouteMapping):
        super().__init__(route, class_instance)
        self.route_mapping = route_mapping

    def get_url(self) -> str:
        url = ""
        if self.parent:
            url = self.route
            parent = self.parent
            while parent is not None:
                url = f"{parent.route}/{url}"
                parent = parent.parent  # type: ignore
        return f"/{url}"


StrOrBytes = Union[str, bytes]
HeaderList = List[Tuple[bytes, bytes]]
TreePart = Union[Branch, Leaf]

Scope = MutableMapping[str, Any]
Message = MutableMapping[str, Any]

Receive = Callable[[], Awaitable[Message]]
Send = Callable[[Message], Awaitable[None]]



import inspect
import json
import mimetypes
import urllib.parse
from collections import namedtuple
from dataclasses import is_dataclass
from io import IOBase, TextIOBase
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

import dacite

from tonberry import content_types
from tonberry.contexted.request import Request
from tonberry.contexted.response import Response
from tonberry.exceptions import FigureItOutLaterException, RouteNotFoundError
from tonberry.expose import _Expose
from tonberry.models import Branch, Leaf, TreePart
from tonberry.util import DataClassEncoder, File, format_data

if TYPE_CHECKING:
    from tonberry.app import App


class Router:
    def __init__(self, app: "App"):
        self.app = app
        self._response: Response = Response()

    async def format_response_body(self, result: Any) -> bytes:
        if isinstance(result, (dict, list)) or is_dataclass(result):
            if not self._response.content_type:
                self._response.content_type = "application/json"
            return json.dumps(result, cls=DataClassEncoder).encode("utf-8")
        if isinstance(result, TextIOBase):
            self.app.app_logger.warning(
                "open().read() is a blocking operation! Use tonberry.File() instead."
            )
            return result.read().encode("utf-8")
        if isinstance(result, IOBase):
            self.app.app_logger.warning(
                "open().read() is a blocking operation! Use tonberry.File() instead."
            )
            return b"".join(result.readlines())
        if isinstance(result, File):
            return await result.read()
        if isinstance(result, str):
            return result.encode("utf-8")
        if isinstance(result, bytes):
            return result
        raise NotImplementedError


class DynamicRouter(Router):
    async def handle_request(self, request: Request, response: Response) -> Response:
        raise NotImplementedError

    async def handle_ws_request(self, request: Request) -> Tuple[Callable, List[Any]]:
        raise NotImplementedError


class MethodRouter(DynamicRouter):
    def __init__(self, app: "App"):
        super().__init__(app)
        # noinspection PyProtectedMember
        self.method_registration = _Expose._registrar
        self._root: Optional[Type] = None
        self._tree: Optional[TreePart] = None

    @property
    def root(self) -> Optional[Type]:
        return self._root

    @root.setter
    def root(self, root: Type) -> None:
        self._root = root
        self._tree = Branch("", root, self.build_tree(root))

    def set_content_type_from_annotation(self, annotation) -> None:  # type: ignore
        if not self._response.content_type and annotation in content_types.ContentTypes:
            self._response.content_type = content_types.CONTENT_TYPE_MAP[annotation]

    async def handle_request(self, request: Request, response: Response) -> Response:
        self._response = response
        response.status = 200
        # noinspection PyTypeHints
        response.body = await self._dispatch(request)  # type: ignore
        return response

    async def handle_ws_request(self, request: Request) -> Tuple[Callable, List[Any]]:
        func = self.get_func(request)
        return func, self.get_args(request)

    async def _dispatch(self, request: Request) -> bytes:
        func = self.get_func(request)
        return await self.call_func(request, func)

    @staticmethod
    async def get_kwargs(request: Request) -> Dict:
        kwargs: Dict[str, Any] = {}
        body = await request.get_body()
        content_type = request.headers["content-type"]
        if content_type is None:
            content_type = ""

        if body:
            if "json" in content_type:
                kwargs.update(json.loads(body))
            if content_type == "application/x-www-form-urlencoded":
                kwargs.update(format_data(urllib.parse.parse_qs(body)))

        if request.query_string:
            kwargs.update(request.query_string)
        return kwargs

    @staticmethod
    def get_args(request: Request) -> List[Any]:
        if request.current_route is not None:
            args: List[Any] = [request.current_route.class_instance]
        else:
            args = []
        # noinspection PyProtectedMember
        args.extend(
            [
                path
                for path in request._unsearched_path.split("/")
                if path and path != "index"
            ]
        )
        return args

    async def call_func(self, request: Request, func: Callable) -> bytes:
        kwargs = await self.get_kwargs(request)
        args = self.get_args(request)

        ArgTypes = namedtuple("ArgTypes", "positional keyword")
        arg_types = ArgTypes(args, {})
        sig = inspect.signature(func)
        params = sig.parameters
        return_annotation = getattr(sig, "return_annotation")
        if return_annotation:
            self.set_content_type_from_annotation(return_annotation)
        var_pos = False
        var_keyword = False
        for name, param in params.items():
            if (
                kwargs
                and hasattr(param, "annotation")
                and is_dataclass(param.annotation)
            ):
                instance = dacite.from_dict(param.annotation, kwargs)
                for key in list(kwargs.keys()):
                    if hasattr(instance, key):
                        del kwargs[key]
                arg_types.positional.append(instance)
            if name in kwargs:
                if param.kind in (
                    inspect.Parameter.POSITIONAL_ONLY,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    inspect.Parameter.VAR_POSITIONAL,
                ):
                    arg_types.positional.append(kwargs.pop(name))
                else:
                    arg_types.keyword[name] = kwargs.pop(name)
            if param.kind == inspect.Parameter.VAR_POSITIONAL:
                var_pos = True
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                var_keyword = True

        for key in list(kwargs.keys()):
            if var_pos:
                arg_types.positional.append(kwargs.pop(key))
            elif var_keyword:
                arg_types.keyword[key] = kwargs.pop(key)

        if func is not None and not kwargs:
            result = await func(*arg_types.positional, **arg_types.keyword)
            return await self.format_response_body(result)

        raise RouteNotFoundError

    def get_func(self, request: Request) -> Callable:
        http_method = request.method
        url_path = request.path
        paths = [urllib.parse.unquote(pth) for pth in url_path.split("/") if pth != ""]
        paths.append("index")
        if self._tree is None:
            raise FigureItOutLaterException
        route = self._tree
        level = 0
        while isinstance(route, Branch):
            if len(paths) <= level:
                raise RouteNotFoundError
            route = self.match_route(paths[level], route.children, http_method)
            level += 1
        if route and isinstance(route, Leaf):
            request._unsearched_path = "/".join(paths[level:])
            request.current_route = route
            if route.route_mapping.func is not None:
                return route.route_mapping.func
            else:
                raise FigureItOutLaterException
        else:
            raise RouteNotFoundError

    @staticmethod
    def match_route(
        current_path: str, routes: List[TreePart], http_method: str
    ) -> TreePart:
        for route in routes:
            if route.route == current_path:
                if isinstance(route, Leaf):
                    if route.route_mapping.http_method == http_method:
                        return route
                elif isinstance(route, Branch):
                    return route
        raise RouteNotFoundError

    def build_tree(self, cls: Type) -> List[TreePart]:
        children: List[TreePart] = []

        for key, value in cls.__class__.__dict__.items():
            if not key.startswith("_"):
                if (
                    key.islower()
                    and hasattr(value, "__class__")
                    and inspect.isclass(value.__class__)
                    and not inspect.ismethod(value)
                    and not inspect.isfunction(value)
                ):
                    childs = self.build_tree(value)
                    branch = Branch(key, value, childs)
                    for child in childs:
                        child.parent = branch
                    children.append(branch)
                elif inspect.ismethod(value) or inspect.isfunction(value):
                    mappings = self.method_registration.get_all_maps_by_func(
                        value.__name__, cls.__class__.__name__
                    )
                    mappings = {mapping for mapping in mappings if mapping}
                    for mapping in mappings:
                        children.append(Leaf(mapping.route, cls, mapping))

        return children


class StaticRouter(Router):
    def __init__(self, app: "App", path_root: Union[str, Path], route: str = ""):
        super().__init__(app)
        self.route = route
        if isinstance(path_root, str):
            path_root = Path(path_root)
        self.path_root = path_root.absolute().resolve()

    async def handle_request(self, request: Request, response: Response) -> Response:
        if request.path.startswith(self.route):
            self._response = response
            response.status = 200
            # noinspection PyTypeHints
            response.body = await self._dispatch(request)  # type: ignore
            return response
        raise RouteNotFoundError

    async def _dispatch(self, request: Request) -> bytes:
        file = (
            self.path_root / request.path.replace(f"{self.route}/", "", 1)
        ).resolve()
        if file.is_file():
            mime_type, encoding = mimetypes.guess_type(file.as_posix())
            if mime_type:
                self._response.content_type = mime_type
            if encoding:
                self._response.headers["Content-Encoding"] = encoding
            file_obj = File(file)
            return await file_obj.read()
        raise RouteNotFoundError



from typing import Any, AsyncGenerator, Dict, Optional, Tuple, Union
from urllib.parse import ParseResult, ParseResultBytes, parse_qs, urlparse

import user_agents
from user_agents.parsers import UserAgent

from tonberry.header import Header
from tonberry.models import Receive, Scope, StrOrBytes, TreePart
from tonberry.util import format_data


class Request:
    def __init__(self, scope: Scope, recieve: Receive):
        self._recieve = recieve
        self.method = str(scope.get("method"))
        self.type = scope.get("type")
        self.http_version = scope.get("http_version")
        self._uri: Union[ParseResult, ParseResultBytes] = urlparse(scope.get("path"))
        self.raw_uri = scope.get("raw_path")
        self.root_path = scope.get("root_path")
        self.client: Tuple[str, str] = scope.get("client")  # type: ignore
        self._query_string = scope.get("query_string")
        self._body: Optional[bytes] = None
        self.headers = Header(scope.get("headers"))
        self.current_route: Optional[TreePart] = None
        self._unsearched_path: str = ""
        self._user_agent: str = ""
        self._scope = scope

    async def stream(self) -> AsyncGenerator[bytes, None]:
        while True:
            chunk = await self._recieve()
            if chunk["type"] == "http.request":
                body = chunk.get("body", b"")
                if body:
                    yield body
                if not chunk.get("more_body", False):
                    break
        yield b""

    async def get_body(self) -> bytes:
        if self._body is None:
            chunks = []
            async for chunk in self.stream():
                chunks.append(chunk)
            self._body = b"".join(chunks)
        return self._body

    @property
    def body(self) -> Optional[bytes]:
        return self._body

    @property
    def path(self) -> str:
        if isinstance(self._uri.path, bytes):
            return self._uri.path.decode("utf-8")
        else:
            return self._uri.path

    @property
    def scheme(self) -> StrOrBytes:
        return self._uri.scheme

    @property
    def netloc(self) -> StrOrBytes:
        return self._uri.netloc

    @property
    def params(self) -> StrOrBytes:
        return self._uri.params

    @property
    def query_string(self) -> Dict:
        return format_data(parse_qs(self._query_string))

    @property
    def raw_query_string(self) -> Any:
        return self._query_string

    @property
    def fragment(self) -> StrOrBytes:
        return self._uri.fragment

    @property
    def username(self) -> Optional[StrOrBytes]:
        return self._uri.username

    @property
    def password(self) -> Optional[StrOrBytes]:
        return self._uri.password

    @property
    def hostname(self) -> Optional[StrOrBytes]:
        return self._uri.hostname

    @property
    def port(self) -> Optional[int]:
        return self._uri.port

    @property
    def user_agent(self) -> UserAgent:
        if not self._user_agent:
            self._user_agent = user_agents.parse(
                self.headers["user-agent"] or "unknown"
            )
        return self._user_agent



from typing import Any, Dict, Hashable
from uuid import UUID


class Session:
    def __init__(self, session_id: UUID, data: Dict = None):
        self.session_id = session_id
        self.data = data or {}

    def get(self, item: Hashable, default: Any = None) -> Any:
        return self.data.get(item, default)

    def __getitem__(self, item: Hashable) -> Any:
        return self.data[item]

    def __setitem__(self, key: Hashable, value: Any) -> None:
        self.data[key] = value

    def __contains__(self, item: Hashable) -> bool:
        return item in self.data


class SessionStore:
    def __init__(self, sessions: Dict[UUID, Session] = None):
        self.sessions = sessions or {}

    def __getitem__(self, item: UUID) -> Session:
        if item not in self.sessions:
            self.sessions[item] = Session(item)
        return self.sessions[item]

    def __setitem__(self, key: UUID, value: Session) -> None:
        self.sessions[key] = value

    def __contains__(self, item: UUID) -> bool:
        return item in self.sessions



from datetime import datetime
from http.cookies import SimpleCookie
from typing import Dict, Optional, Union

from tonberry.models import HeaderList, StrOrBytes


class Header:
    def __init__(self, header: HeaderList = None):
        if header is None:
            header = []
        self._header = header
        self._header_attrs: Dict[str, str] = {}
        self.decode()

    def __contains__(self, item: str) -> bool:
        if item in self._header_attrs:
            return True
        return False

    def __getitem__(self, item: str) -> Optional[str]:
        if item in self._header_attrs:
            return self._header_attrs[item]
        return None

    def __setitem__(self, key: str, value: Union[str, int]) -> None:
        self._header_attrs[key] = str(value)

    def __delitem__(self, key: str) -> None:
        del self._header_attrs[key]

    def encode(self) -> HeaderList:
        raw_header = []
        for header_key, attr in self._header_attrs.items():
            key = bytes(header_key, "utf-8")
            if attr is None:
                value = b""
            else:
                value = bytes(attr, "utf-8")
            raw_header.append((key, value))
        return raw_header

    def decode(self) -> None:
        for key, value in self._header:
            self._header_attrs[key.decode("utf-8")] = value.decode("utf-8")

    def set_cookie(
        self,
        key: str,
        morsel: str,
        path: StrOrBytes = None,
        domain: StrOrBytes = None,
        secure: bool = False,
        expires: datetime = None,
        max_age: int = None,
        comment: str = None,
        version: str = None,
    ) -> None:
        if "Set-Cookie" not in self:
            self._header_attrs["Set-Cookie"] = ""
        cookie: SimpleCookie = SimpleCookie()
        cookie[key] = morsel
        if path is not None:
            cookie[key]["path"] = path
        if domain is not None:
            cookie[key]["domain"] = domain
        if secure:
            cookie[key]["secure"] = True
        if expires is not None:
            cookie[key]["expires"] = expires.strftime("%a, %d %b %Y %H:%M:%S")
        if max_age is not None:
            cookie[key]["max-age"] = max_age
        if comment is not None:
            cookie[key]["comment"] = comment
        if version is not None:
            cookie[key]["version"] = version
        self._header_attrs["Set-Cookie"] += str(cookie)

    def get_cookie(self, name: str = None) -> Optional[Union[SimpleCookie, str]]:
        cookie: SimpleCookie = SimpleCookie()
        if "cookie" in self._header_attrs:
            cookie_dough = self._header_attrs["cookie"]
            cookie_dough = cookie_dough.replace("Set-Cookie: ", "")
            cookie.load(cookie_dough)
            if name is not None:
                if name in cookie:
                    morsel = cookie.get(name)
                    if morsel is not None:
                        return morsel.value
            return cookie
        return None



from dataclasses import dataclass

from tonberry import File, expose, quick_start, request, session, create_app
from tonberry.content_types import ApplicationJson, TextHTML, TextPlain
from tonberry.exceptions import HTTPRedirect


@dataclass
class Request1:
    thing1: int
    thing2: str


@dataclass
class Request2:
    thing1: str
    thing2: str


class ChildOne:
    @expose.get
    async def index(self) -> TextPlain:
        return "Child 1"

    @expose.get
    async def sit(self) -> TextPlain:
        return "Child 1 Sat Down"

    @expose.post
    async def stand_up(self) -> TextPlain:
        return "Child 1 Stood Up"

    @expose.post
    async def stuff_with_json(self, request1: Request1) -> TextPlain:
        return f"Child 1 has {request1.thing1} {request1.thing2}"

    @expose.post
    async def stuff_with_urlencoded(self, request2: Request2) -> TextPlain:
        return f"Child 1 has {request2.thing1} {request2.thing2}"


class ChildsChild:
    @expose.get
    async def index(self) -> TextPlain:
        return "Child Child Index"

    @expose.get
    async def something(self) -> TextPlain:
        url = request.current_route.get_url()
        print(url)
        return "something"

    @expose.get
    async def shout(self, something: str = "stuff") -> TextPlain:
        return f"AHHHHH {something}"


class ChildTwo:
    childschild = ChildsChild()

    def __init__(self, child: int) -> None:
        self.child = child

    @expose.get
    async def index(self) -> TextHTML:
        self.child += 1
        return File("test.html")

    @expose.get
    async def sit(self, name: str) -> TextPlain:
        return f"Child 2 named {name} Sat Down"

    @expose.get("up")
    async def stand_up(self) -> TextPlain:
        return "Child 2 Stood Up"


class Root:
    child_one = ChildOne()
    child_two = ChildTwo(0)

    @expose.get
    async def index(self) -> TextPlain:
        return "Hello, how are you?"

    @expose.get
    async def hey(self) -> TextHTML:
        num = session.get("num", 0) + 1
        print(num)
        session["num"] = num
        return File("test.html")

    @expose.get
    async def what(self, thing, num) -> TextPlain:
        return f"Hello {thing} {num}"

    @expose.post
    async def what(self, thing, num) -> TextPlain:
        return f"Go away {thing} {num}"

    @expose.get
    async def hello(self) -> TextPlain:
        return "GET Tonberry says Hello World"

    @expose.post
    async def hello(self) -> TextPlain:
        return "POST Tonberry says Hello World"

    @expose.get
    async def getjson(self) -> ApplicationJson:
        thing = self.do_a_thing(2)
        return {"hello": "world", "thing": thing}

    @staticmethod
    def do_a_thing(num: int) -> int:
        return num * 2

    @expose.get
    def i_redirect(self):
        raise HTTPRedirect("/getjson")


app = create_app(Root)


if __name__ == "__main__":
    quick_start(Root, "127.0.0.1", 8000)



from tonberry.header import Header


class Body:
    chunk_size = 1024

    def __init__(self, data: bytes):
        self.data = data
        self._position = 0

    def __aiter__(self) -> "Body":
        return self

    async def __anext__(self) -> bytes:
        data = self.data[self._position : self._position + self.chunk_size]
        if not data:
            self._position = 0
            raise StopAsyncIteration
        self._position += self.chunk_size
        return data


class Response:
    def __init__(self, headers: Header = None):
        self._body = Body(b"")
        self.headers: Header = headers or Header()
        self.status: int = 200
        self._content_type: str = ""
        self.timeout = 60

    @property
    def body(self) -> Body:
        return self._body

    @body.setter
    def body(self, data: bytes) -> None:
        self._body.data = data

    @property
    def content_type(self) -> str:
        return self._content_type

    @content_type.setter
    def content_type(self, value: str) -> None:
        self._content_type = value
        self.headers["content-type"] = value



class TonberryException(Exception):
    pass


class HTTPRedirect(TonberryException):
    def __init__(self, route: str, code: int = 307):
        self.route = route
        self.code = code


class HTTPError(TonberryException):
    pass


class RouteNotFoundError(TonberryException):
    pass


class FigureItOutLaterException(TonberryException):
    pass


class WebSocketError(TonberryException):
    pass


class WebSocketDisconnect(TonberryException):
    pass



import asyncio
import json
from concurrent.futures import ThreadPoolExecutor
from dataclasses import asdict, is_dataclass
from functools import partial
from pathlib import Path
from typing import Any, Dict, Union

from tonberry.models import StrOrBytes


class DataClassEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Dict[str, Any]:
        if is_dataclass(obj):
            return asdict(obj)
        return json.JSONEncoder.default(self, obj)


def decode_bytes_to_str(value: StrOrBytes) -> str:
    """
    Will convert any byte strings to UTF-8 strings
    """
    if isinstance(value, bytes):
        value = value.decode("utf-8")
    return value


def format_data(kwargs: Dict) -> Dict:
    """
    De-lists any single-item lists and scrubs the values

    So {b'foo': [b'bar']} would become {'foo': 'bar'}
    """
    new_kwargs = {}
    for key, value in kwargs.items():
        if isinstance(value, list):
            for index, val in enumerate(value):
                value[index] = decode_bytes_to_str(val)
        if len(value) == 1:
            value = decode_bytes_to_str(value[0])
        new_kwargs[decode_bytes_to_str(key)] = value
    return new_kwargs


class File:
    def __init__(self, path: Union[str, Path]) -> None:
        if isinstance(path, str):
            path = Path(path)
        self.path = path
        self.chunk_size = 64

    async def read(self) -> bytes:
        chunks = []
        loop = asyncio.get_event_loop()
        pool = ThreadPoolExecutor()
        open_file = self.path.open("rb")
        read_func = partial(open_file.read, self.chunk_size)
        while True:
            chunk = await loop.run_in_executor(pool, read_func)
            if not chunk:
                break
            chunks.append(chunk)
        return b"".join(chunks)



import sys
from typing import Callable, Optional, Union

from tonberry.exceptions import FigureItOutLaterException
from tonberry.models import Methods, RouteMapping


class _Expose:
    _registrar = Methods()

    @classmethod
    def _new_registrar(cls) -> None:
        cls._registrar = Methods()

    @classmethod
    def _register(
        cls,
        func: Callable,
        http_method: str,
        route: Optional[Union[str, Callable]] = None,
        frame_index: int = 3,
    ) -> None:
        if route is None or not isinstance(route, str):
            route = func.__name__
        # noinspection PyProtectedMember
        parents = sys._getframe(frame_index).f_locals
        if "__qualname__" in parents and "__module__" in parents:
            module_name = parents["__module__"]
            qual_name = parents["__qualname__"]
        else:
            raise FigureItOutLaterException
        cls._registrar.get(http_method).mappings.append(
            RouteMapping(route, func, module_name, qual_name, http_method)
        )

    @classmethod
    def get(cls, route: Union[str, Callable] = None) -> Callable:
        def wrapper(wrapped_func: Callable, frame_index: int = 2) -> Callable:
            cls._register(wrapped_func, "GET", route, frame_index)
            return wrapped_func

        if callable(route):
            func = route
            route = None
            return wrapper(func, 3)
        else:
            return wrapper

    @classmethod
    def post(cls, route: Union[str, Callable] = None) -> Callable:
        def wrapper(wrapped_func: Callable, frame_index: int = 2) -> Callable:
            cls._register(wrapped_func, "POST", route, frame_index)
            return wrapped_func

        if callable(route):
            func = route
            route = None
            return wrapper(func, 3)
        else:
            return wrapper

    @classmethod
    def put(cls, route: Union[str, Callable] = None) -> Callable:
        def wrapper(wrapped_func: Callable, frame_index: int = 2) -> Callable:
            cls._register(wrapped_func, "PUT", route, frame_index)
            return wrapped_func

        if callable(route):
            func = route
            route = None
            return wrapper(func, 3)
        else:
            return wrapper

    @classmethod
    def delete(cls, route: Union[str, Callable] = None) -> Callable:
        def wrapper(wrapped_func: Callable, frame_index: int = 2) -> Callable:
            cls._register(wrapped_func, "DELETE", route, frame_index)
            return wrapped_func

        if callable(route):
            func = route
            route = None
            return wrapper(func, 3)
        else:
            return wrapper

    @classmethod
    def patch(cls, route: Union[str, Callable] = None) -> Callable:
        def wrapper(wrapped_func: Callable, frame_index: int = 2) -> Callable:
            cls._register(wrapped_func, "PATCH", route, frame_index)
            return wrapped_func

        if callable(route):
            func = route
            route = None
            return wrapper(func, 3)
        else:
            return wrapper

    @classmethod
    def head(cls, route: Union[str, Callable] = None) -> Callable:
        def wrapper(wrapped_func: Callable, frame_index: int = 2) -> Callable:
            cls._register(wrapped_func, "HEAD", route, frame_index)
            return wrapped_func

        if callable(route):
            func = route
            route = None
            return wrapper(func, 3)
        else:
            return wrapper

    @classmethod
    def options(cls, route: Union[str, Callable] = None) -> Callable:
        def wrapper(wrapped_func: Callable, frame_index: int = 2) -> Callable:
            cls._register(wrapped_func, "OPTIONS", route, frame_index)
            return wrapped_func

        if callable(route):
            func = route
            route = None
            return wrapper(func, 3)
        else:
            return wrapper

    @classmethod
    def websocket(cls, route: Union[str, Callable] = None) -> Callable:
        def wrapper(wrapped_func: Callable, frame_index: int = 2) -> Callable:
            cls._register(wrapped_func, "WEBSOCKET", route, frame_index)
            return wrapped_func

        if callable(route):
            func = route
            route = None
            return wrapper(func, 3)
        else:
            return wrapper



from _contextvars import ContextVar
from typing import Any, Hashable


class ContextVarManager:
    def __init__(self, name: str):
        object.__setattr__(self, "_context_var", ContextVar(name))

    @property
    def _context(self) -> Any:
        return self._context_var.get()

    def get(self, item: Hashable, default: Any) -> Any:
        return self._context.get(item, default)

    def __getitem__(self, item: Hashable) -> Any:
        return self._context[item]

    def __setitem__(self, key: Hashable, value: Any) -> None:
        self._context[key] = value

    def __delitem__(self, key: Hashable) -> None:
        del self._context[key]

    def __setattr__(self, key: str, value: Any) -> None:
        setattr(self._context, key, value)

    def __getattr__(self, item: str) -> Any:
        return getattr(self._context, item)


def set_context_var(context_var_manager: Any, var: Any) -> None:
    # noinspection PyProtectedMember
    context_var_manager._context_var.set(var)



from typing import IO, TypeVar

from tonberry import File

TextHTML = TypeVar("TextHTML", str, bytes, IO, File)
TextPlain = TypeVar("TextPlain", str, bytes, IO, File)
TextCss = TypeVar("TextCss", str, bytes, IO, File)
TextCsv = TypeVar("TextCsv", str, bytes, IO, File)
TextJavaScript = TypeVar("TextJavaScript", str, bytes, IO, File)
TextXml = TypeVar("TextXml", str, bytes, IO, File)

ApplicationJavaArchive = TypeVar("ApplicationJavaArchive", bytes, IO, File)
ApplicationEdiX12 = TypeVar("ApplicationEdiX12", str, bytes, IO, File)
ApplicationEdiFact = TypeVar("ApplicationEdiFact", str, bytes, IO, File)
ApplicationJavaScript = TypeVar("ApplicationJavaScript", str, bytes, IO, File)
ApplicationOctetStream = TypeVar("ApplicationOctetStream", str, bytes, IO, File)
ApplicationOgg = TypeVar("ApplicationOgg", bytes, IO, File)
ApplicationPdf = TypeVar("ApplicationPdf", bytes, IO, File)
ApplicationXHtmlXml = TypeVar("ApplicationXHtmlXml", str, bytes, IO, File)
ApplicationXShockwaveFlash = TypeVar("ApplicationXShockwaveFlash", bytes, IO, File)
ApplicationJson = TypeVar("ApplicationJson", str, bytes, dict, list, IO, File)
ApplicationLdJson = TypeVar("ApplicationLdJson", str, bytes, IO, File)
ApplicationXml = TypeVar("ApplicationXml", str, bytes, IO, File)
ApplicationZip = TypeVar("ApplicationZip", bytes, IO, File)
ApplicationXWWWFormUrlEncoded = TypeVar(
    "ApplicationXWWWFormUrlEncoded", str, bytes, IO, File
)

ApplicationVndAndroidPackageArchive = TypeVar(
    "ApplicationVndAndroidPackageArchive", bytes, IO, File
)
ApplicationVndOasisOpenDocumentText = TypeVar(
    "ApplicationVndOasisOpenDocumentText", bytes, IO, File
)
ApplicationVndOasisOpenDocumentSpreadsheet = TypeVar(
    "ApplicationVndOasisOpenDocumentSpreadsheet", bytes, IO, File
)
ApplicationVndOasisOpenDocumentPresentation = TypeVar(
    "ApplicationVndOasisOpenDocumentPresentation", bytes, IO, File
)
ApplicationVndOasisOpenDocumentGraphics = TypeVar(
    "ApplicationVndOasisOpenDocumentGraphics", bytes, IO, File
)
ApplicationVndMsExcel = TypeVar("ApplicationVndMsExcel", bytes, IO, File)
ApplicationVndOpenXmlFormatsOfficeDocumentSpreadsheetmlSheet = TypeVar(
    "ApplicationVndOpenXmlFormatsOfficeDocumentSpreadsheetmlSheet", bytes, IO, File
)
ApplicationVndMsPowerpoint = TypeVar("ApplicationVndMsPowerpoint", bytes, IO, File)
ApplicationVndOpenXmlFormatsOfficeDocumentPresentationmlPresentation = TypeVar(
    "ApplicationVndOpenXmlFormatsOfficeDocumentPresentationmlPresentation",
    bytes,
    IO,
    File,
)
ApplicationMsWord = TypeVar("ApplicationMsWord", bytes, IO,)
ApplicationVndOpenXmlFormatsOfficeDocumentWordProcessingmlDocument = TypeVar(
    "ApplicationVndOpenXmlFormatsOfficeDocumentWordProcessingmlDocument",
    bytes,
    IO,
    File,
)
ApplicationVndMozillaXulXml = TypeVar("ApplicationVndMozillaXulXml", bytes, IO, File)

AudioMpeg = TypeVar("AudioMpeg", bytes, IO, File)
AudioXMsWma = TypeVar("AudioXMsWma", bytes, IO, File)
AudioVndRnRealAudio = TypeVar("AudioVndRnRealAudio", bytes, IO, File)
AudioXWav = TypeVar("AudioXWav", bytes, IO, File)

ImageGif = TypeVar("ImageGif", bytes, IO, File)
ImageJpeg = TypeVar("ImageJpeg", bytes, IO, File)
ImagePng = TypeVar("ImagePng", bytes, IO, File)
ImageTiff = TypeVar("ImageTiff", bytes, IO, File)
ImageVndMicrosoftIcon = TypeVar("ImageVndMicrosoftIcon", bytes, IO, File)
ImageXIcon = TypeVar("ImageXIcon", bytes, IO, File)
ImageVndDjvu = TypeVar("ImageVndDjvu", bytes, IO, File)
ImageSvgXml = TypeVar("ImageSvgXml", bytes, IO, File)

MultipartMixed = TypeVar("MultipartMixed", str, bytes, IO, File)
MultipartAlternative = TypeVar("MultipartAlternative", str, bytes, IO, File)
MultipartRelated = TypeVar("MultipartRelated", str, bytes, IO, File)
MultipartFormData = TypeVar("MultipartFormData", str, bytes, IO, File)

VideoMpeg = TypeVar("VideoMpeg", bytes, IO, File)
VideoMp4 = TypeVar("VideoMp4", bytes, IO, File)
VideoQuicktime = TypeVar("VideoQuicktime", bytes, IO, File)
VideoXMsWmv = TypeVar("VideoXMsWmv", bytes, IO, File)
VideoXMsVideo = TypeVar("VideoXMsVideo", bytes, IO, File)
VideoXFlv = TypeVar("VideoXFlv", bytes, IO, File)
VideoWebm = TypeVar("VideoWebm", bytes, IO, File)

ContentTypes = [
    TextHTML,
    TextPlain,
    TextCss,
    TextCsv,
    TextJavaScript,
    TextXml,
    ApplicationJavaArchive,
    ApplicationEdiX12,
    ApplicationEdiFact,
    ApplicationJavaScript,
    ApplicationOctetStream,
    ApplicationOgg,
    ApplicationPdf,
    ApplicationXHtmlXml,
    ApplicationXShockwaveFlash,
    ApplicationJson,
    ApplicationLdJson,
    ApplicationXml,
    ApplicationZip,
    ApplicationXWWWFormUrlEncoded,
    AudioMpeg,
    AudioXMsWma,
    AudioVndRnRealAudio,
    AudioXWav,
    ImageGif,
    ImageJpeg,
    ImagePng,
    ImageTiff,
    ImageVndMicrosoftIcon,
    ImageXIcon,
    ImageVndDjvu,
    ImageSvgXml,
    MultipartMixed,
    MultipartAlternative,
    MultipartRelated,
    MultipartFormData,
    VideoMpeg,
    VideoMp4,
    VideoQuicktime,
    VideoXMsWmv,
    VideoXMsVideo,
    VideoXFlv,
    VideoWebm,
    ApplicationVndAndroidPackageArchive,
    ApplicationVndOasisOpenDocumentText,
    ApplicationVndOasisOpenDocumentSpreadsheet,
    ApplicationVndOasisOpenDocumentPresentation,
    ApplicationVndOasisOpenDocumentGraphics,
    ApplicationVndMsExcel,
    ApplicationVndOpenXmlFormatsOfficeDocumentSpreadsheetmlSheet,
    ApplicationVndMsPowerpoint,
    ApplicationVndOpenXmlFormatsOfficeDocumentPresentationmlPresentation,
    ApplicationMsWord,
    ApplicationVndOpenXmlFormatsOfficeDocumentWordProcessingmlDocument,
    ApplicationVndMozillaXulXml,
]

CONTENT_TYPE_MAP = {
    TextHTML: "text/html",
    TextPlain: "text/plain",
    TextCss: "text/css",
    TextCsv: "text/csv",
    TextJavaScript: "text/javascript",
    TextXml: "text/xml",
    ApplicationJavaArchive: "application/java-archive",
    ApplicationEdiX12: "application/EDI-X12",
    ApplicationEdiFact: "application/EDIFACT",
    ApplicationJavaScript: "application/javascript",
    ApplicationOctetStream: "application/octet-stream",
    ApplicationOgg: "application/ogg",
    ApplicationPdf: "application/pdf",
    ApplicationXHtmlXml: "application/xhtml+xml",
    ApplicationXShockwaveFlash: "application/x-shockwave-flash",
    ApplicationJson: "application/json",
    ApplicationLdJson: "application/ld+json",
    ApplicationXml: "application/xml",
    ApplicationZip: "application/zip",
    ApplicationXWWWFormUrlEncoded: "application/x-www-form-urlencoded",
    AudioMpeg: "audio/mpeg",
    AudioXMsWma: "audio/x-ms-wma",
    AudioVndRnRealAudio: "audio/vnd.rn-realaudio",
    AudioXWav: "audio/x-wav",
    ImageGif: "image/gif",
    ImageJpeg: "image/jpeg",
    ImagePng: "image/png",
    ImageTiff: "image/tiff",
    ImageVndMicrosoftIcon: "image/vnd.microsoft.icon",
    ImageXIcon: "image/x-icon",
    ImageVndDjvu: "image/vnd.djvu",
    ImageSvgXml: "image/svg+xml",
    MultipartMixed: "multipart/mixed",
    MultipartAlternative: "multipart/alternative",
    MultipartRelated: "multipart/related",
    MultipartFormData: "multipart/form-data",
    VideoMpeg: "video/mpeg",
    VideoMp4: "video/mp4",
    VideoQuicktime: "video/quicktime",
    VideoXMsWmv: "video/x-ms-wmv",
    VideoXMsVideo: "video/x-msvideo",
    VideoXFlv: "video/x-flv",
    VideoWebm: "video/webm",
    ApplicationVndAndroidPackageArchive: "application/vnd.android.package-archive",
    ApplicationVndOasisOpenDocumentText: "application/vnd.oasis.opendocument.text",
    ApplicationVndOasisOpenDocumentSpreadsheet: "application/vnd.oasis.opendocument."
    "spreadsheet",
    ApplicationVndOasisOpenDocumentPresentation: "application/vnd.oasis.opendocument."
    "presentation",
    ApplicationVndOasisOpenDocumentGraphics: "application/vnd.oasis.opendocument."
    "graphics",
    ApplicationVndMsExcel: "application/vnd.ms-excel",
    ApplicationVndOpenXmlFormatsOfficeDocumentSpreadsheetmlSheet: "application/vnd."
    "openxmlformats-officedocument.spreadsheetml.sheet",
    ApplicationVndMsPowerpoint: "application/vnd.ms-powerpoint",
    ApplicationVndOpenXmlFormatsOfficeDocumentPresentationmlPresentation: "application/"
    "vnd.openxmlformats-officedocument.presentationml.presentation",
    ApplicationMsWord: "application/msword",
    ApplicationVndOpenXmlFormatsOfficeDocumentWordProcessingmlDocument: "application/"
    "vnd.openxmlformats-officedocument.wordprocessingml.document",
    ApplicationVndMozillaXulXml: "application/vnd.mozilla.xul+xml",
}



from typing import TYPE_CHECKING, Type

from tonberry.context_var_manager import ContextVarManager
from tonberry.contexted.request import Request
from tonberry.contexted.response import Response
from tonberry.contexted.session import Session
from tonberry.expose import _Expose
from tonberry.util import File
from tonberry.websocket import WebSocket

if TYPE_CHECKING:
    from tonberry.app import App
    from tonberry.routers import Router

expose = _Expose

# noinspection PyTypeChecker
request: Request = ContextVarManager("request")  # type: ignore
# noinspection PyTypeChecker
response: Response = ContextVarManager("response")  # type: ignore
# noinspection PyTypeChecker
session: Session = ContextVarManager("session")  # type: ignore
# noinspection PyTypeChecker
websocket: WebSocket = ContextVarManager("websocket")  # type: ignore


def create_app(root: Type = None) -> "App":
    from tonberry.app import App

    app_instance = App()
    if root is not None:
        app_instance.routers[0].root = root()  # type: ignore
    return app_instance


def quick_start(root: Type, host: str = "localhost", port: int = 8080) -> None:
    import uvicorn

    quick_app = create_app()
    # noinspection PyTypeHints
    quick_app.routers[0].root = root()  # type: ignore

    uvicorn.run(quick_app, host=host, port=port, log_level="info", access_log=False)



import asyncio
import traceback
from typing import TYPE_CHECKING

from tonberry import response as response_context
from tonberry.context_var_manager import set_context_var
from tonberry.contexted.request import Request
from tonberry.contexted.response import Response
from tonberry.exceptions import HTTPError, HTTPRedirect
from tonberry.models import Receive, Scope, Send
from tonberry.websocket import WebSocket

if TYPE_CHECKING:
    from tonberry.app import App


class Handler:
    def __init__(self, app: "App", scope: Scope):
        self.app = app
        self.scope = scope

    async def __call__(self, recieve: Receive, send: Send) -> None:
        raise NotImplementedError


class HTTPHandler(Handler):
    def __init__(self, app: "App", scope: Scope):
        super().__init__(app, scope)

    async def __call__(self, recieve: Receive, send: Send) -> None:
        request = Request(self.scope, recieve)
        try:
            await self.handle_request(request, send)
        except HTTPError as err:
            response = Response()
            response.status = err.args[0]
            # noinspection PyTypeHints
            response.body = str(err.args[0]).encode("utf-8")  # type: ignore
            set_context_var(response_context, response)
            self.app.http_access_logger.error()
            await self.handle_exception(response, send)
        except HTTPRedirect as err:
            response = Response()
            response.status = err.code
            response.headers["Location"] = err.route
            set_context_var(response_context, response)
            self.app.http_access_logger.info()
            await self.handle_exception(response, send)
        except Exception as err:
            response = Response()
            # noinspection PyTypeHints
            response.body = traceback.format_exc().encode("utf-8")  # type: ignore
            response.status = 500
            set_context_var(response_context, response)
            self.app.http_access_logger.error()
            self.app.app_logger.exception(err)
            await self.handle_exception(response, send)

    async def handle_request(self, request: Request, send: Send) -> None:
        session_id = self.app.get_session_id(request)
        response: Response = await self.app.handle_request(request)
        if request.headers.get_cookie("TBSESSIONID") is None:
            response.headers.set_cookie(
                "TBSESSIONID",
                str(session_id),
                path=request.path,
                domain=request.hostname,
            )
        try:
            await asyncio.wait_for(
                self.respond(send, response), timeout=response.timeout
            )
        except asyncio.TimeoutError:
            pass

    async def handle_exception(self, response: Response, send: Send) -> None:
        try:
            await asyncio.wait_for(
                self.respond(send, response), timeout=response.timeout
            )
        except asyncio.TimeoutError:
            pass

    @staticmethod
    async def respond(send: Send, response: Response) -> None:
        await send(
            {
                "type": "http.response.start",
                "status": response.status,
                "headers": response.headers.encode(),
            }
        )
        async for data in response.body:
            await send({"type": "http.response.body", "body": data, "more_body": True})
        await send({"type": "http.response.body", "body": b"", "more_body": False})


class WebSocketHandler(Handler):
    def __init__(self, app: "App", scope: Scope):
        super().__init__(app, scope)

    async def __call__(self, recieve: Receive, send: Send) -> None:
        request = Request(self.scope, recieve)
        request.method = "WEBSOCKET"
        websocket = WebSocket(self.app, self.scope, recieve, send)
        await self.app.handle_ws_request(websocket, request)


class LifespanHandler(Handler):
    def __init__(self, app: "App", scope: Scope):
        super().__init__(app, scope)

    async def __call__(self, recieve: Receive, send: Send) -> None:
        message = await recieve()
        if message["type"] == "lifespan.startup":
            try:
                self.app.startup()
            except Exception as err:
                await send({"type": "lifespan.startup.failed", "message": str(err)})
            else:
                await send({"type": "lifespan.startup.complete"})
        elif message["type"] == "lifespan.shutdown":
            try:
                self.app.shutdown()
            except Exception as err:
                await send({"type": "lifespan.shutdown.failed", "message": str(err)})
            else:
                await send({"type": "lifespan.shutdown.complete"})



from pathlib import Path
from typing import Callable, Generator, List, Union
from uuid import UUID, uuid4

from websockets import ConnectionClosedError, ConnectionClosedOK

from tonberry import request as request_context
from tonberry import response as response_context
from tonberry import session as session_context
from tonberry import websocket as websocket_context
from tonberry.context_var_manager import set_context_var
from tonberry.contexted.request import Request
from tonberry.contexted.response import Response
from tonberry.contexted.session import Session, SessionStore
from tonberry.exceptions import (
    HTTPError,
    RouteNotFoundError,
    WebSocketDisconnect,
    WebSocketError,
)
from tonberry.handlers import HTTPHandler, LifespanHandler, WebSocketHandler
from tonberry.loggers import (
    create_app_logger,
    create_http_access_logger,
    create_websocket_access_logger,
)
from tonberry.models import Receive, Scope, Send
from tonberry.routers import DynamicRouter, MethodRouter, Router, StaticRouter
from tonberry.websocket import WebSocket


class App:
    def __init__(self, routers: List[Router] = None):
        self.routers = routers or [MethodRouter(self)]
        self.http_access_logger = create_http_access_logger()
        self.websocket_access_logger = create_websocket_access_logger()
        self.app_logger = create_app_logger()
        self.sessions = SessionStore()
        self.startup_functions: List[Callable] = []
        self.shutdown_functions: List[Callable] = []

    async def __call__(self, scope: Scope, recieve: Receive, send: Send) -> None:
        if scope["type"] == "http":
            handler = HTTPHandler(self, scope)
        elif scope["type"] == "websocket":
            handler = WebSocketHandler(self, scope)  # type: ignore
        elif scope["type"] == "lifespan":
            handler = LifespanHandler(self, scope)  # type: ignore
        else:
            raise RuntimeError(
                f"{scope['type']} is either not a standard ASGI scope type or is not "
                f"yet implimented"
            )
        await handler(recieve, send)

    @property
    def static_routes(self) -> Generator[StaticRouter, None, None]:
        return (router for router in self.routers if isinstance(router, StaticRouter))

    @property
    def dynamic_routes(self) -> Generator[DynamicRouter, None, None]:
        return (router for router in self.routers if isinstance(router, DynamicRouter))

    def add_router(self, router: Router) -> None:
        self.routers.append(router)

    def add_static_route(self, path_root: Union[str, Path], route: str = "") -> None:
        self.add_router(StaticRouter(self, path_root, route))

    def on_startup(self, func: Callable) -> None:
        self.startup_functions.append(func)

    def on_shutdown(self, func: Callable) -> None:
        self.shutdown_functions.append(func)

    def startup(self) -> None:
        for func in self.startup_functions:
            func()

    def shutdown(self) -> None:
        for func in self.shutdown_functions:
            func()

    async def handle_request(self, request: Request) -> Response:
        set_context_var(request_context, request)
        response = Response()
        set_context_var(response_context, response)
        for s_router in self.static_routes:
            try:
                response = await s_router.handle_request(request, response)
                break
            except RouteNotFoundError:
                pass
        else:  # no break
            for d_router in self.dynamic_routes:
                try:
                    response = await d_router.handle_request(request, response)
                    break
                except RouteNotFoundError:
                    pass
            else:  # no break
                raise HTTPError(404)
        self.http_access_logger.info()
        return response

    async def handle_ws_request(self, websocket: WebSocket, request: Request) -> None:
        set_context_var(request_context, request)
        set_context_var(websocket_context, websocket)
        for router in self.dynamic_routes:
            try:
                func, args = await router.handle_ws_request(request)
                await websocket.accept()
                self.websocket_access_logger.info("Opened")
                try:
                    await func(*args)
                except (WebSocketDisconnect, ConnectionClosedOK):
                    self.websocket_access_logger.info("Disconnected")
                except ConnectionClosedError:
                    self.websocket_access_logger.error("Disconnected unexpectedly")
                break
            except RouteNotFoundError:
                pass
        else:  # no break
            raise WebSocketError

    def get_session_id(self, request: Request) -> UUID:
        sesson_cookie = request.headers.get_cookie("TBSESSIONID")
        if sesson_cookie is not None:
            session_id = UUID(str(sesson_cookie))
            session = self.sessions[session_id]
            set_context_var(session_context, session)
            return session_id
        else:
            return self.create_session()

    def create_session(self) -> UUID:
        new_session_id = uuid4()
        self.sessions[new_session_id] = Session(new_session_id)
        set_context_var(session_context, self.sessions[new_session_id])
        return new_session_id



from logging import INFO, Formatter, Logger, LogRecord, StreamHandler

from tonberry import request, response


class TonberryLogger(Logger):
    def __init__(self, name: str, level: int):
        super().__init__(name, level)

    def info(self, msg: str = "", *args, **kwargs) -> None:  # type: ignore
        super().info(msg, *args, **kwargs)

    def error(self, msg: str = "", *args, **kwargs) -> None:  # type: ignore
        super().error(msg, *args, **kwargs)


class TonberryHTTPLogger(TonberryLogger):
    # noinspection PyTypeHints
    def makeRecord(  # type: ignore
        self,
        name,
        level,
        fn,
        lno,
        msg,
        args,
        exc_info,
        func=None,
        extra=None,
        sinfo=None,
    ) -> LogRecord:
        record = super().makeRecord(
            name, level, fn, lno, msg, args, exc_info, func, extra, sinfo
        )
        record.client_ip: str = request.client[0]  # type: ignore
        record.client_port: str = request.client[1]  # type: ignore
        record.path: str = request.path  # type: ignore
        record.http_method: str = request.method  # type: ignore
        record.http_version: str = request.http_version  # type: ignore
        record.status: int = response.status  # type: ignore
        return record


class TonberryWebsocketLogger(TonberryLogger):
    # noinspection PyTypeHints
    def makeRecord(  # type: ignore
        self,
        name,
        level,
        fn,
        lno,
        msg,
        args,
        exc_info,
        func=None,
        extra=None,
        sinfo=None,
    ) -> LogRecord:
        record = super().makeRecord(
            name, level, fn, lno, msg, args, exc_info, func, extra, sinfo
        )
        record.client_ip: str = request.client[0]  # type: ignore
        record.client_port: str = request.client[1]  # type: ignore
        record.path: str = request.path  # type: ignore
        return record


def create_http_access_logger() -> TonberryHTTPLogger:
    logger = TonberryHTTPLogger("tb_http_access", INFO)
    handler = StreamHandler()
    formatter = Formatter(
        "{asctime} {levelname} {client_ip}:{client_port} {http_method} "
        "HTTP/{http_version} {status} {path} {message}",
        style="{",
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def create_websocket_access_logger() -> TonberryWebsocketLogger:
    logger = TonberryWebsocketLogger("tb_ws_access", INFO)
    handler = StreamHandler()
    formatter = Formatter(
        "{asctime} {levelname} {client_ip}:{client_port} WebSocket {path} {message}",
        style="{",
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def create_app_logger() -> TonberryLogger:
    logger = TonberryLogger("tb_app", INFO)
    handler = StreamHandler()
    formatter = Formatter("{asctime} {levelname} {message}", style="{")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger
